;;; sml-simplify
;;;
;;; ICFP 2001 Programming Contest
;;; (c) Matthieu Villeneuve


;;; General utilities

(defun char-to-string (char)
  (make-string 1 :initial-element char))

(defun read-string-until (stream delimiters)
  (do ((result (make-string 0))
       (c (read-char stream nil) (read-char stream nil)))
      ((or (null c) (member c delimiters)) result)
    (setf result (concatenate 'string result (char-to-string c)))))

(defmacro while (test &rest body)
  `(do ()
       ((not ,test))
     ,@body))


;;; Blank characters

(defconstant +blank-chars+ '(#\Space #\Newline #\Linefeed #\Tab))

(defun char-is-blank (char)
  (member char +blank-chars+ :test #'equal))

(defun string-is-blank-char (string)
  (and (= (length string) 1)
	   (char-is-blank (aref string 0))))


;;; SML Constants & Utils

(defconstant +sml-tag-tt+ "TT")
(defconstant +sml-color-tags+ '("r" "g" "b" "c" "m" "y" "k" "w"))
(defconstant +sml-nonredundant-tags+ '("U"))

(defun is-color-tag (string)
  (member string +sml-color-tags+ :test #'equal))

(defun is-nonredundant-tag (string)
  (member string +sml-nonredundant-tags+ :test #'equal))

(defun sml-tree-size (tree)
  (cond ((null tree) 0)
		((stringp tree) 1)
		(t (1+ (apply #'+ (mapcar #'sml-tree-size (cdr tree)))))))


;;; SML Decoding

(defun decode-sml-subtree (label is)
  (let ((children '())
		(token "")
		(tree '()))
	(do ((c (read-char is nil) (when (null tree) (read-char is nil))))
		((not (null tree)) tree)
	  (cond ((null c)
			 (when (> (length token) 0)
			   (setf children (append children (list token))))
			 (setf tree (cons label children)))
			((eq c #\<)
			 (when (> (length token) 0)
			   (setf children (append children (list token)))
			   (setf token ""))
			 (let ((tag (read-string-until is '(#\>))))
			   (if (eq (aref tag 0) #\/)
				   (if (equal (subseq tag 1) label)
					   (setf tree (cons label children))
					   (error "Tags do not match"))
				   (let ((subtree (decode-sml-subtree tag is)))
					 (setf children (append children (list subtree)))))))
			((member c +blank-chars+)
			 (let* ((new-char (char-to-string c))
					(new-children (if (> (length token) 0)
									  (list token new-char)
									  (list new-char))))
			   (setf children (append children new-children))
			   (setf token "")))
			(t
			 (let ((new-char (char-to-string c)))
			   (setf token (concatenate 'string token new-char))))))))

(defun decode-sml-tree (is)
  (decode-sml-subtree "PL" is))

(defun decode-sml-file (input-file)
  (let ((input-path (make-pathname :name input-file)))
    (with-open-file (is input-path)
      (decode-sml-tree is))))


;;; SML Encoding

(defun output-string (string os)
  (format os "~A" string))

(defun output-open-tag (tag os)
  (format os "<~A>" tag))

(defun output-close-tag (tag os)
  (format os "</~A>" tag))

(defun encode-sml-tree (tree os)
  (let ((tag (car tree)))
	(output-open-tag tag os)
	(dolist (elt (cdr tree))
	  (cond ((stringp elt)
			 (output-string elt os))
			((consp elt)
			 (encode-sml-tree elt os))))
	(output-close-tag tag os)))

(defun encode-sml-tree-list (list os)
  (cond ((null list) nil)
		((stringp (car list))
		 (output-string (car list) os)
		 (encode-sml-tree-list (cdr list) os))
		(t 
		 (encode-sml-tree (car list) os)
		 (encode-sml-tree-list (cdr list) os))))


;;; Filter: unify blank characters as spaces
;;;  <B> \t bar</B> => <B>   bar</B>

(defun filter-unify-sp-tree-list (list)
  (cond ((null list) nil)
		(t (let ((elt (car list))
				 (rest (filter-unify-sp-tree-list (cdr list))))
			 (cond ((stringp elt)
					(if (string-is-blank-char elt)
						(cons " " rest)
						(cons elt rest)))
				   ((consp elt)
					(cons (filter-unify-sp-tree elt) rest))
				   (t (error "Not a valid SML tree")))))))

(defun filter-unify-sp-tree (tree)
  (if (equal (car tree) +sml-tag-tt+)
	  tree
	  (cons (car tree) (filter-unify-sp-tree-list (cdr tree)))))


;;; Filter: remove duplicate spaces
;;;  <I>foo <B> bar </B> </I> => <I>foo <B>bar </B></I>

(defun filter-remove-dup-sp-tree-list-aux (list last-sp)
  (cond ((null list) (values nil last-sp))
		(t (let ((elt (car list)))
			 (cond ((and (stringp elt) (equal elt " "))
					(multiple-value-bind (result sp)
					    (filter-remove-dup-sp-tree-list-aux (cdr list) t)
					  (if last-sp
						  (values result sp)
						  (values (cons (car list) result) sp))))
				   ((stringp elt)
					(multiple-value-bind (result sp)
					    (filter-remove-dup-sp-tree-list-aux (cdr list) nil)
					  (values (cons (car list) result) sp)))
				   (t 
					(multiple-value-bind (result sp)
						(filter-remove-dup-sp-tree-aux (car list) last-sp)
					  (multiple-value-bind (result2 sp2)
						  (filter-remove-dup-sp-tree-list-aux (cdr list) sp)
						(values (cons result result2) sp2)))))))))

(defun filter-remove-dup-sp-tree-aux (tree last-sp)
  (cond ((null tree) (values nil last-sp))
		((equal (car tree) +sml-tag-tt+) (values tree nil))
		(t (multiple-value-bind (result sp)
			   (filter-remove-dup-sp-tree-list-aux (cdr tree) last-sp)
			 (values (cons (car tree) result) sp)))))

(defun filter-remove-dup-sp-tree (tree)
  (filter-remove-dup-sp-tree-aux tree nil))


;;; Filter: remove empty subtrees
;;;  foo<B></B>bar => foobar

(defun filter-prune-tree-list (list)
  (cond ((null list) nil)
		(t (let ((elt (car list))
				 (rest (filter-prune-tree-list (cdr list))))
			 (cond ((stringp elt)
					(cons elt rest))
				   ((consp elt)
					(let ((pruned-elt (filter-prune-tree elt)))
					  (if (null pruned-elt)
						  rest
						  (cons pruned-elt rest))))
				   (t (print elt)))))))

(defun filter-prune-tree (tree)
  (cond ((or (null tree) (null (cdr tree))) nil)
		(t (let ((children (filter-prune-tree-list (cdr tree))))
			 (cond ((null children) nil)
				   (t (cons (car tree) children)))))))


;;; Filter: group consecutive similar trees
;;;  <B>foo</B><B>bar</B> => <B>foobar</B>

(defun filter-group-similar-tree-list (list)
  (labels ((group-tree-list (list)
			 (while (and (not (or (null list) (null (cdr list))))
						 (consp (car list))
						 (consp (cadr list))
						 (equal (caar list) (caadr list)))
			   (let ((third (cddr list)))
				 (setf list (cons (append (car list) (cdadr list))
								  third))))
			 (cond ((or (null list) (null (cdr list))) list)
				   (t (cons (car list) (group-tree-list (cdr list)))))))
	(cond ((null list) nil)
		  (t (setf list (mapcar #'filter-group-similar-tree list))
			 (group-tree-list list)))))

(defun filter-group-similar-tree (tree)
  (cond ((null tree) nil)
		((stringp tree) tree)
		(t (cons (car tree) (filter-group-similar-tree-list (cdr tree))))))


;;; Filter: rearrange color trees
;;;  <r>foo</r><b>bar</b><r>baz</r> => <r>foo<b>bar</b>baz</r>

(defun filter-group-color-tree-list (list)
  (labels ((group-tree-list (list)
			 (while (and (not (or (null list) 
								  (null (cdr list))
								  (null (cddr list))))
						 (consp (car list)) (is-color-tag (caar list))
						 (consp (cadr list)) (is-color-tag (caadr list))
						 (consp (caddr list)) (is-color-tag (caaddr list))
						 (equal (caar list) (caaddr list)))
			   (let ((fourth (cdddr list)))
				 (setf list (cons (append (car list)
										  (list (cadr list))
										  (cdaddr list))
								  fourth))))
			 (cond ((or (null list) (null (cdr list))) list)
				   (t (cons (car list) (group-tree-list (cdr list)))))))
	(cond ((null list) nil)
		  (t (setf list (mapcar #'filter-group-color-tree list))
			 (group-tree-list list)))))

(defun filter-group-color-tree (tree)
  (cond ((null tree) nil)
		((stringp tree) tree)
		(t (cons (car tree) (filter-group-color-tree-list (cdr tree))))))


;;; Filter: remove redundant tags
;;;  <B><B><B>foo</B></B></B> => <B>foo</B>

(defun filter-rem-redundant-tree-list (list label)
  (labels ((group-tree-list (list)
			 (if (and (not (null list))
					  (consp (car list))
					  (equal (caar list) label))
				 (setf list (append (cdar list) (cdr list))))
			 (cond ((null list) nil)
				   (t (cons (car list) (group-tree-list (cdr list)))))))
	(cond ((null list) nil)
		  (t (setf list (mapcar #'filter-rem-redundant-tree list))
			 (if (is-nonredundant-tag label)
				 list
				 (group-tree-list list))))))

(defun filter-rem-redundant-tree (tree)
  (cond ((null tree) nil)
		((stringp tree) tree)
		(t (cons (car tree) 
				 (filter-rem-redundant-tree-list (cdr tree) (car tree))))))


;;; Main function

(defun optimize (filename)
  (let* ((tree (decode-sml-file filename))
		 (size (sml-tree-size tree)))
	(do ((old-size 0))
		((= size old-size) (values))
	  (setf old-size size)
	  (setf tree (filter-rem-redundant-tree
				  (filter-group-color-tree
				   (filter-group-similar-tree
					(filter-prune-tree
					 (filter-remove-dup-sp-tree
					  (filter-unify-sp-tree tree)))))))
	  (setf size (sml-tree-size tree)))
    (encode-sml-tree-list (cdr tree) *standard-output*))
  (values))
